"""

"""

import importlib.resources as pkg_resources
import json
from typing import Any, Dict

import jugmt.schema


def get_schema() -> Dict[str, Any]:
    """
    Load the stage1 JSON Schema from package datafrom return it as a dictionary.

    Returns:
        Dict[str, Any]: The JSON Schema as a dictionary.
    """

    with pkg_resources.open_text(jugmt.schema, "stage1.schema.json") as content:
        return json.load(content)



"""
Encapsulation of Document meta-data, .docx

Assumptions
===========

* Every NVMe specification document has a TOC with a List of figures
"""

import json
import re
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import docx
from docx.table import Table
from jinja2 import Environment, PackageLoader, select_autoescape
from jsonschema import Draft202012Validator, validate

from jugmt.schema.checker import get_schema

REGEX_FIGURE_IDENTIFIER = (
    r"(?P<title>Figure\s+(?P<figure_nr>\d+):(?P<description>.*)\s)(?P<page_nr>\d+)"
)

TEMPLATE_HTML = "document.html.jinja2"


def table_to_dict(table: Table) -> List[List[Dict[str, Any]]]:
    """
    Returns a JSON serializable form of the given docx.table.Table.

    The table is represented as a list-of-lists, thus, it can be irregular. This is
    intentional, as the tables coming from the input can be irregular, thus, we need the
    representation to do a lint check for this.
    """

    rows = []
    for row in table.rows:
        cols = []
        for cell in row.cells:
            cols.append(
                {
                    "text": str(cell.text),
                    "tables": [
                        table_to_dict(nested_table) for nested_table in cell.tables
                    ],
                }
            )
        rows.append(cols)

    return rows


@dataclass
class Figure(object):
    """
    A figure as represented in the NVMe specification documents.
    """

    figure_nr: int  # Figure as numbered in the specification document
    page_nr: int  # The page, in the document, that the figure starts on
    title: str  # The entire figure title
    description: str  # The part of figure title without the "Fig X:" prefix

    table: Optional[List[List[Dict[str, Any]]]] = None


class Document(object):
    """
    Wrapper of the docx.Document with additional path meta-data and figures with tabular
    data converted to a JSON serializable format
    """

    def __init__(self, path: Path):
        self.path = path
        self.docx = docx.Document(path)
        self.figures: List[Any] = []

        self.json_str: Optional[str] = None

    def extract_figures(self):
        """
        Populate self.figures with figures, and their associated tabular data when
        available, found in self.docx
        """

        # Convert all tables to serializable format, and when doing so, built a map of
        #
        #   figure-description => figure
        #
        # By doing so, then lookup for the matching the figure is much faster
        tables = {}  # map "description" => "table"
        for docx_table in self.docx.tables:
            key = ("".join(docx_table.rows[0].cells[0].text.split(":")[1:])).strip()
            tables[key] = table_to_dict(docx_table)

        prev = cur = None
        for paragraph in self.docx.paragraphs:
            cur = paragraph.style.name

            # We exit early to avoid scanning the entire document, since we known that
            # once we are looking at a "table of figures" paragraph, then once we see
            # one that is not, then no more will arrive
            if prev == "table of figures" and cur != "table of figures":
                break
            prev = cur

            # Skip paragraphs before "table of figures"
            if paragraph.style.name != "table of figures":
                continue

            # Check whether the paragraph is a reference to a figure
            text = paragraph.text.strip()
            match = re.match(REGEX_FIGURE_IDENTIFIER, text)
            if not match:
                continue

            # Construct figure and lookup whether a table exists for it
            figure = Figure(
                figure_nr=int(match.group("figure_nr")),
                page_nr=int(match.group("page_nr")),
                title=match.group("title").strip(),
                description=match.group("description").strip(),
            )
            figure.table = tables.get(figure.description, None)

            self.figures.append(figure)

    def to_html(self) -> str:
        """Returns the document as a HTML-formatted string"""

        env = Environment(
            loader=PackageLoader("jugmt", "templates"),
            autoescape=select_autoescape(["html", "xml"]),
        )
        template = env.get_template(TEMPLATE_HTML)

        return template.render(document=self)

    def to_json(self) -> str:
        """Returns the document as a JSON-formatted string"""

        if not self.json_str:
            self.json_str = json.dumps(
                {
                    "path": str(self.path),
                    "figures": [asdict(fig) for fig in self.figures],
                },
                indent="\t",
            )

        return self.json_str

    def validate(self, level: int) -> bool:
        """Validate the document using the schema and the Draft202012 Validator"""

        if self.json_str is None:
            self.json_str = self.to_json()

        if level == 1:
            return Draft202012Validator(get_schema()).is_valid(self.json_str)
        elif level == 2:
            return validate(
                instance=json.loads(self.json_str),
                schema=get_schema(),
                cls=Draft202012Validator,
            )

        return False



"""
Command-Line Interface
======================

Produces .html and .json when given .docx documents
"""

import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path

from jugmt.document import Document


def parse_args() -> Namespace:
    """Return command-line arguments"""

    parser = ArgumentParser(
        description="Extract table information from .docx documents"
    )
    parser.add_argument(
        "document", nargs="+", type=Path, help="Path to one or more .docx document(s)"
    )
    parser.add_argument(
        "--output",
        type=Path,
        help="Directory where the output will be saved",
        default=Path("output"),
    )
    parser.add_argument(
        "--validate",
        type=int,
        default=1,
        choices=[0, 1, 2],
        help="0: skip validation; 1; true/false; 2: trace of errors",
    )

    return parser.parse_args()


def main() -> int:
    """CLI entrypoint"""

    args = parse_args()
    args.output.mkdir(parents=True, exist_ok=True)

    for path in args.document:
        document = Document(path)
        document.extract_figures()

        (args.output / document.path.stem).with_suffix(".html").write_text(
            document.to_html()
        )
        json_str = document.to_json()
        (args.output / document.path.stem).with_suffix(".json").write_text(json_str)

        if args.validate == 0:
            continue

        document.validate(args.validate)

    return 0


if __name__ == "__main__":
    sys.exit(main())



from subprocess import run
from pathlib import Path
import re
import jugmt

REGEX_PYTHON_SLOC=r"^python:\s+(?P<nlines>\d+)\s+\((?P<percent>\d+\.\d+)%\)$"

def project_sloc_jinja2():
    """Count number of lines in jinja2 templates"""

    sloc = 0
    for path in Path(jugmt.__file__).parent.rglob("*.jinja2"):
        with path.open() as content:
            sloc += sum(1 for line in content)

    return sloc

def project_sloc_python():
    """Count number of lines of Python code, returning (nlines, npercent)"""
    
    path = Path(jugmt.__file__).parent
    
    result = run(
        ["sloccount", f"{path}"],
        capture_output=True,
        text=True,
    )

    nlines = 0
    npercent = 0
    for line in result.stdout.splitlines():
        match = re.match(REGEX_PYTHON_SLOC, line)
        if match:
            nlines = int(match.group("nlines"))
            npercent = float(match.group("percent"))

    return nlines, npercent
            
def test_weight():
    """Test that the jugmt source do not grow too large"""
    
    python_sloc, python_percent = project_sloc_python()
    
    assert python_percent == 100, "This is a Python project; expecting 100% Python code"
    assert python_sloc < 300, "Too much Python code; simplify the code"

    jinja2_sloc = project_sloc_jinja2()
    assert jinja2_sloc < 150, "Too much jinja2 template-code; simplify the code"

    assert jinja2_sloc, "The codebase uses jinja2 templates, so this must be >0"
    assert python_percent, "The codebase is Python; must have percent Python code >0"
    assert python_sloc, "The codebase is Python; must have line of code > 0"


#!/usr/bin/env python3
from pathlib import Path
from subprocess import run


def test_cli_tool(tmp_path):

    paths = list(Path("documents").resolve().glob("*.docx"))
    assert len(paths) > 0, "No documents/*.docx available for testing"

    for path in paths:
        result = run(
            ["jugmt", "--output", str(tmp_path), f"{path}"],
            capture_output=True,
            text=True,
        )

        assert result.returncode == 0, f"CLI tool failed with error: {result.stderr}"
